/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 01 - exercise 18 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>

#define MAX_FLIGHTS 150

/* structure declaration */
typedef struct {
  char code[6];
  char depCity[4]; 
  char arrCity[4];
  float depTime;
  float arrTime;
} flight_t;

/* function prototypes */
int flight_read (flight_t *flights, char *name);
void connection_search (flight_t *flights, int nf, flight_t request);
void info_write (flight_t flight);

/* 
 *  main program
 */
int main(int argc, char *argv[]) {
  flight_t flights[MAX_FLIGHTS], request;
  int nf;

  if (argc < 2) {
    printf("Error: missing parameter.\n");
    printf("Run as: %s <input_file>.\n", argv[0]);
    return 1;
  }

  /* load all fligths */
  nf = flight_read(flights, argv[1]);
  if (nf == 0) {
    return 1;
  }

  /* input requested connection */
  printf("Introduce the departure city: ");
  scanf("%s", request.depCity);
  printf("Introduce the arrival city: ");
  scanf("%s", request.arrCity);
  printf("Introduce the departure time: ");
  scanf("%f", &request.depTime);

  /* search for the request */
  connection_search(flights, nf, request);
  return 0;
}

/* 
 *  read the flights list from file
 */
int flight_read (flight_t *flights, char *name) {
  char line[100];
  FILE *fp;
  int i=0;

  fp = fopen(name, "r");
  if (fp == NULL) {
    printf("Error opening the input file.\n");
    return 0;
  }

  while (fgets(line, 100, fp)!=NULL && i<MAX_FLIGHTS) {
    sscanf(line, "%s%s%s%f%f", flights[i].code, 
                  flights[i].depCity, flights[i].arrCity, 
                  &flights[i].depTime, &flights[i].arrTime);
    i++;
  }

  fclose(fp);
  return i;
}

/* 
 *  search all flights satisfying the request
 */
void connection_search (flight_t *flights, int nf, flight_t request) {
  int i, j, found;

  for (i=0; i<nf; i++) {
    if (strcmp(request.depCity, flights[i].depCity)==0 && 
        request.depTime<=flights[i].depTime) {
      if (strcmp(request.arrCity, flights[i].arrCity)==0) {
        /* found a direct flight: print info */
        printf("Direct flight:\n");
        info_write(flights[i]);
      } else {
        /* it might exist a non direct flight: search */
        found = 0;
        for (j=0; j<nf && !found; j++) {
          if (strcmp(flights[j].depCity, flights[i].arrCity)==0 && 
              flights[j].depTime>=flights[i].arrTime &&
              strcmp(request.arrCity, flights[j].arrCity)==0) {
            /* found a non direct flight: print info */
            printf("Flight with one stop:\n");
            info_write(flights[i]);
            info_write(flights[j]);
            found = 1; /* print only one connection */
          }
        }
      }
    }
  }
}

/* 
 *  print flight information
 */
void info_write (flight_t flight) {
  printf("%s ", flight.code);
  printf("%s %s ", flight.depCity, flight.arrCity);
  printf("%2.2f %2.2f\n", flight.depTime, flight.arrTime);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
