/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 03 - exercise 03 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define MAX_ID    5+1
#define MAX_NAME 20+1

/* structure declaration */
typedef struct {
  char id[MAX_ID];
  char surname[MAX_NAME];
  char name[MAX_NAME];
  int total_amount;
  int num_donations;
} donor_t;

/* function prototypes */
donor_t *donor_read (int *num_ptr);
void donation_read (donor_t *donors, int n);
int id_search (donor_t *donors, char id[MAX_ID+1], int n);
void result_display (donor_t *donors, int n);

/*
 *  main program
 */
int main (void) {
  donor_t *donors;
  int n;

  donors = donor_read(&n);
  donation_read(donors, n);
  result_display(donors, n);

  free(donors);
  return EXIT_SUCCESS;
}

/*
 *  load the donor list into a dynamic array
 */
donor_t *donor_read (int *num_ptr) {
  char id[MAX_ID], name[MAX_NAME], surname[MAX_NAME];
  donor_t *donors;
  int i, n;
  FILE *fp;

  fprintf(stdout,"Donors file name: ");
  scanf("%s", name);

  /* compute the number of rows */
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr,"File open error (file=%s).\n", name);
    exit(EXIT_FAILURE);
  }
  n = 0;
  while (fscanf(fp, "%s %*s %s %s", id, surname, name) != EOF) {
    n++;
  }
  fclose (fp);

  /* allocate the dynamic array */
  donors = (donor_t *)malloc(n * sizeof(donor_t));
  if (donors == NULL) {
    fprintf(stderr,"Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  /* parse the file contents */
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<n; i++) {
    fscanf(fp, "%s %*s %s %s", id, surname, name);
    strcpy(donors[i].id, id);
    strcpy(donors[i].name, name);
    strcpy(donors[i].surname, surname);
    donors[i].total_amount = 0;
    donors[i].num_donations = 0;
  }
  fclose(fp);

  *num_ptr = n;

  return donors;
}

/*
 *  parse the donation file; update the data structure
 */
void donation_read (donor_t *donors, int n) {
  char name[MAX_NAME], id[MAX_ID];
  int i, amount;
  FILE *fp;

  fprintf(stdout,"Donations file name: ");
  scanf("%s", name);
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  while (fscanf(fp, "%s %d", id, &amount) != EOF) {
    /* update the "donation" amount */
    i = id_search(donors, id, n);
    if (i >= 0) {
      donors[i].total_amount += amount;
      donors[i].num_donations++;
    }
  }

  fclose (fp);

  return;
}

/*
 *  search an "id" into the donor array; return its index
 */
int id_search (donor_t *donors, char id[MAX_ID+1], int n) {
  int i;

  for (i=0; i<n; i++) {
    if (strcmp(donors[i].id, id) == 0) {
      return i;
    }
  }

  return -1;
}

/*
 *  output the required results
 */
void result_display (donor_t *donors, int n) {
  int i;

  for (i=0; i<n; i++) {
    fprintf(stdout, "%s %s ", donors[i].id, donors[i].name);
    fprintf(stdout, "%s: %d", donors[i].surname, donors[i].num_donations);
    fprintf(stdout, " donation(s) - %d cc\n", donors[i].total_amount);
  }

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
