/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 03 - exercise 04 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#define DEAD	0
#define ALIVE	1
#define CHILD	2

#define DIM    19
#define MAX    50

/* structure declaration */
typedef struct {
   char DNA[DIM+1];
   int state;
} organism_t;

/*
 *  parse the input file, return the organism population array
 */
organism_t *input_read (int *n, char *name) {
  organism_t *population;
  FILE* fp;
  int i;

  fp = fopen(name, "r");
  if (fp == NULL){ 
    fprintf(stderr,"File open error.\n");
    exit(EXIT_FAILURE);
  }

  fscanf(fp, "%d", n);
  population = (organism_t *)malloc((*n) * sizeof(organism_t)); 
  if (population == NULL){ 
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<*n; i++) {
    fscanf(fp, "%s", population[i].DNA);
    population[i].state = ALIVE;
  }
  fclose(fp);

  return population;
}

/*
 *  generate a new organism from two parents
 */
void merge (organism_t *population, int i, int j) {
  int k;

  for (k=0; k<DIM; k++) {
    if (population[i].DNA[k] > population[j].DNA[k]) {
      population[i].DNA[k] = population[j].DNA[k];
    }
  }
  population[i].state = CHILD;
  population[j].state = DEAD;
}

/*
 *  compare two organisms "char by char"
 */
int likeness (organism_t *population, int i, int j) {
  int k, count=0;

  for (k=0; k<DIM; k++) {
    if (population[i].DNA[k] == population[j].DNA[k]) {
      count++;
    }
  }

  return count;
}

/*
 *  find the two organisms with maximum likeness
 */
int likeness_max (organism_t *population, int n, int *p1, int *p2) {
  int i, j, like, max=0;

  for (i=0; i<n-1; i++) {
    for (j=i+1; j<n; j++) {
      if (population[i].state==ALIVE && population[j].state==ALIVE) {
        like = likeness(population, i, j);
        if (like > max) {
          *p1 = i;
          *p2 = j;
          max = like;
        }
      }
    }
  }
  return max;
}

/*
 *  evolve the current organism generation; 
 *  return 1 if no new organism can be generated
 */
int evolve (organism_t *population, int n) {
  int i, j, endFlag=1;

  while (likeness_max (population, n, &i, &j) != 0) {
    merge(population, i, j);
    endFlag = 0;
  }

  return endFlag;
}

/*
 *  main program
 */
int main (int argc, char *argv[]) {
  organism_t *population;
  int i, n, generation=1;
  FILE* fp;

  if (argc < 3) {
    fprintf(stderr, "Error: missing parameter.\n");
    fprintf(stderr, "Run as: %s <organism_file> <evolution_file>\n", argv[0]);
    return EXIT_FAILURE;
  }

  population = input_read (&n, argv[1]);

  fp = fopen(argv[2], "w");
  if (fp == NULL) {
    fprintf(stderr, "File open error.\n");
    return EXIT_FAILURE;
  }

  while (evolve(population, n) != 1) {
    /* output a generation */
    fprintf(fp, "Generation %d:\n", generation++);
    for (i=0; i<n; i++) {
      if (population[i].state != DEAD) {
        fprintf(fp, "%s\n", population[i].DNA);
        population[i].state = ALIVE;
      }
    }
  }
  fclose(fp);
  
  free(population);
  return EXIT_SUCCESS;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
