/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 03 - exercise 05 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

/* structure declaration */
typedef struct {
  int x1, y1, x2, y2;
  int used1, used2;
} segment_t;

/* function prototypes */
segment_t *load(char *, int *);
void connect(segment_t *, int);
void check(segment_t *, int);

/*
 *  main program
 */
int main (int argc, char *argv[]) {
  segment_t *segments;
  int n;

  segments = load (argv[1], &n);
  connect (segments, n);
  check (segments, n);

  free(segments);

  return EXIT_SUCCESS;
}

/*
 *  load the input file contents
 */
segment_t *load(char *name, int *n) {
  int i, x1, y1, x2, y2;
  segment_t *segments;
  FILE* fp;

  fp = fopen(name, "r");
  if (fp == NULL){ 
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  fscanf(fp, "%d", n);
  segments = (segment_t *)malloc(*n * sizeof(segment_t)); 
  if (segments == NULL) { 
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  i = 0;
  while (i<*n && fscanf(fp,"%d %d %d %d", &x1, &y1, &x2, &y2)!=EOF) {
    segments[i].x1 = x1;
    segments[i].y1 = y1;
    segments[i].x2 = x2;
    segments[i].y2 = y2;
    segments[i].used1 = 0;
    segments[i].used2 = 0;
    i++;
  }

  *n = i;

  fclose(fp);

  return segments;
}

/*
 *  search for connected segments
 */
void connect(segment_t *segments, int n) {
  int i, j;

  /* for each segment ... */
  for (i=0; i<n-1; i++) {
    /* ... check if one of the next can be connected ... */
    for (j=i+1; j<n; j++) {
      if (segments[i].x1==segments[j].x1 && segments[i].y1==segments[j].y1) {
        segments[i].used1++;
        segments[j].used1++;
      }
      if (segments[i].x1==segments[j].x2 && segments[i].y1==segments[j].y2) {
        segments[i].used1++;
        segments[j].used2++;
      }
      if (segments[i].x2==segments[j].x1 && segments[i].y2==segments[j].y1) {
        segments[i].used2++;
        segments[j].used1++;
      }
      if (segments[i].x2==segments[j].x2 && segments[i].y2==segments[j].y2) {
        segments[i].used2++;
        segments[j].used2++;
      }
    }
  }

  return;
}

/*
 *  check connection status of all segments
 */
void check(segment_t *segments, int n) {
  int i, disconnected, overconnected;

  disconnected = 0;
  overconnected = 0;

  for (i=0; i<n; i++) {
    fprintf(stdout, "segment %2d %2d %2d %2d - ",
      segments[i].x1, segments[i].y1, segments[i].x2, segments[i].y2);
    fprintf(stdout, "left extreme: ");
    if (segments[i].used1 == 0) {
      fprintf(stdout, "disconnected ");
      disconnected++;
    } else if (segments[i].used1 == 1) {
      fprintf(stdout, "OK           ");
    } else {
      fprintf(stdout, "overconnected");
      overconnected++;
    }
    fprintf(stdout, "; right extreme: ");
    if (segments[i].used2 == 0) {
      fprintf(stdout, "disconnected\n");
      disconnected++;
    } else if (segments[i].used2 == 1) {
      fprintf(stdout, "OK\n");
    } else {
      fprintf(stdout, "overconnected\n");
      overconnected++;
    }
  }

  if (disconnected > 2) {
    fprintf(stdout, "Broken line with disconnected segments\n");
  }
  if (overconnected > 0) {
    fprintf(stdout, "Broken line with over-connected segments\n");
  }
  if (disconnected<=2 && overconnected==0) {
    fprintf(stdout, "Single broken line\n");
  }
  
  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
