/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 03 - exercise 06 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

/* function prototypes */
int **matrix_read(char *, int *, int *);
int **product_compute(int **, int, int, int **, int, int);
void matrix_write(char *, int **, int, int);
void matrix_quit(int **, int);

/*
 *  main program
 */
int main (int argc, char *argv[]) {
  int **m1, **m2, **m3;
  int r1, c1, r2, c2;

  if (argc < 4) {
    fprintf(stderr, "Error: missing parameter.\n");
    fprintf(stderr, "Run as: %s <m1_file> <m2_file> <m3_file>\n", argv[0]);
    return EXIT_FAILURE;
  }

  m1 = matrix_read(argv[1], &r1, &c1);
  m2 = matrix_read(argv[2], &r2, &c2);
  if (c1 != r2) {
    fprintf(stderr, "Error: incompatible dimensions.\n");
    return EXIT_FAILURE;
  }
  m3 = product_compute(m1, r1, c1, m2, r2, c2);
  matrix_write(argv[3], m3, r1, c2);

  matrix_quit(m1, r1);
  matrix_quit(m2, r2);
  matrix_quit(m3, r1);
  return EXIT_SUCCESS;
}

/* 
 *  matrix input
 */
int **matrix_read (char *name, int *r, int *c) {
  int i, j, **m;
  FILE *fp;

  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "File open error (file=%s).\n", name);
    exit(EXIT_FAILURE);
  }

  fscanf(fp, "%d %d", r, c);
  m = (int **)malloc((*r) * sizeof(int *));
  if (m == NULL) {
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<*r; i++) {
    m[i] = (int *)malloc((*c) * sizeof(int));
    if (m[i] == NULL) {
      fprintf(stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    for (j=0; j<*c; j++) {
      fscanf(fp, "%d", &m[i][j]);
    }
  }

  fclose(fp);
  return m;
}

/* 
 *  compute the matrix product
 */
int **product_compute (int **m1, int r1, int c1, int **m2, int r2, int c2) {
  int i, j, k, **m3;

  m3 = (int **)malloc(r1 * sizeof(int *));
  if (m3 == NULL) {
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<r1; i++) {
    m3[i] = (int *)malloc(c2 * sizeof(int));
    if (m3[i] == NULL) {
      fprintf(stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    for (j=0; j<c2; j++) {
      m3[i][j] = 0;
      for (k=0; k<c1; k++) {
        m3[i][j] += m1[i][k] * m2[k][j];
      }
    }
  }
  return m3;
}

/* 
 *  matrix output
 */
void matrix_write (char *name, int **m, int r, int c) {
  int i, j;
  FILE *fp;

  fp = fopen(name, "w");
  if (fp == NULL) {
    fprintf(stderr, "File open error (file=%s).\n", name);
    exit(EXIT_FAILURE);
  }

  fprintf(fp, "%d %d\n", r, c);
  for (i=0; i<r; i++) {
    for (j=0; j<c; j++) {
      fprintf(fp, "%d ", m[i][j]);
    }
    fprintf(fp, "\n");
  }
  fclose(fp);
}

/* 
 *  matrix deallocation
 */
void matrix_quit (int **m, int r) {
  int i;

  for (i=0; i<r; i++) {
    free(m[i]);
  }
  free(m);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
