/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 03 - exercise 07 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define MAX_LINE 60+1

/* function prototypes */
char **input_read (char *name, int *num_ptr);
void string_sort (char **strings, int num);
void output_write (char *name, char **strings, int num);

/*
 *  main program
 */
int main (int argc, char *argv[]) {
  char **strings;
  int i, num;
 
  if (argc < 3) {
    fprintf(stderr, "Error: missing parameter.\n");
    fprintf(stderr, "Run as: %s <input_file> <output_file>\n", argv[0]);
    return EXIT_FAILURE;
  }

  strings = input_read(argv[1], &num);
  string_sort(strings, num);
  output_write(argv[2], strings, num);

  for (i=0; i<num; i++) {
    free(strings[i]);
  }
  free(strings);

  return EXIT_SUCCESS;
} 

/*
 *  load the input file contents
 */
char **input_read (char *name, int *num_ptr) {
  char word[MAX_LINE], **strings;
  int n, i;
  FILE *fp;

  /* count the words */
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "File open error %s.\n", name);
    exit(EXIT_FAILURE);
  }

  i = 0;
  while (fscanf(fp, "%s", word) != EOF) {
    i++;
  }
  fclose(fp);
  n = *num_ptr = i;

  /* allocate data structure */
  strings = (char **)malloc(n * sizeof(char *));
  if (strings == NULL) {
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  /* save the words in the array */
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "File open error (file=%s).\n", name);
    exit(EXIT_FAILURE);
  }

  for (i=0; i<n; i++) {
    fscanf(fp, "%s", word);
    strings[i] = (char *)malloc((strlen(word)+1) * sizeof(char));
    if (strings[i] == NULL)  {
      fprintf(stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    strcpy(strings[i], word);
  }
  fclose(fp);

  return strings;
}

/*
 *  insertion sort on (dynamic) strings
 */
void string_sort (char **strings, int num) {
  int i, j;
  char *ptr;

  for (i=1; i<num; i++) {
    ptr = strings[i];
    j = i;
    while (--j>=0 && strcmp(ptr, strings[j])<0) {
      strings[j+1] = strings[j];
    }
    strings[j+1] = ptr;
  }
}

/*
 *  write the output file
 */
void output_write (char *name, char **strings, int num) {
  FILE *fp;
  int i;

  fp = fopen(name, "w");
  if (fp == NULL) {
    fprintf(stderr, "File open error (file=%s).\n", name);
    exit(EXIT_FAILURE);
  }

  for (i=0; i<num; i++) {
    fprintf(fp, "%s\n", strings[i]);
  }
  fclose(fp);
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
