/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 03 - exercise 09 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define MAX    61
#define DAY   365

/* structure declaration */
typedef struct {
  char *surname;
  char *name;
  int in;
  int out;
} request_t;

/* function prototypes */
request_t *file_read(int *);
int roomNumber(request_t *, int);
void reservation_make(request_t *, int, int * [], int);
void file_write(int * [], int, request_t *);

/*
 *  main program
 */
int main (void) {
  int *reservations[DAY];
  request_t *requests;
  int i, n, roomNum;

  requests = file_read(&n);
  roomNum = roomNumber(requests, n);
  reservation_make(requests, n, reservations, roomNum);
  file_write(reservations, roomNum, requests);
  
  for (i=0; i<n; i++) {
    free(requests[i].name);
    free(requests[i].surname);
  }
  free(requests);
  for (i=0; i<DAY; i++) {
    free(reservations[i]);
  }

  return EXIT_SUCCESS;
}

/*
 *  load the request file
 */
request_t *file_read (int *n_ptr) {
  request_t *requests;
  char word[MAX];
  FILE *fp;
  int i;

  fprintf(stdout, "Input file name: ");
  scanf("%s", word);
  fp = fopen(word, "r");
  if (fp == NULL) {
    fprintf(stderr, "File open error (file=%s).\n", word);
    exit(EXIT_FAILURE);
  } 
  
  fscanf(fp, "%d", n_ptr);
  requests = (request_t *)malloc((*n_ptr) * sizeof(request_t));
  if (requests == NULL) {
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<*n_ptr; i++) {
    fscanf(fp, "%s", word);
    requests[i].surname = strdup(word);
    fscanf(fp, "%s", word);
    requests[i].name = strdup(word);
    if (requests[i].surname==NULL || requests[i].name==NULL) {
      fprintf(stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    fscanf(fp, "%d %d", &requests[i].in, &requests[i].out);
  }
  fclose (fp);

  return requests;
}

/*
 *  compute the room number
 */
int roomNumber(request_t *requests, int n)
{
  int i, j, roomNum=0, tmp;
  
  /* for each day ... */
  for (i=1; i<=DAY; i++) {
    tmp = 0;

    /* ... and for each reservation ... */  
    for (j=0; j<n; j++) {
      /* ... check the day */
      if (i>=requests[j].in && i<requests[j].out) {
        tmp++;
      }
    }
    
    if (tmp > roomNum) {
      roomNum = tmp;
    }
  }

  return roomNum;
}

/*
 *  assign rooms to reservations
 */
void reservation_make (
  request_t *requests, int n, int *reservations[], int roomNum
) {
  int i, j, k, in, out, stop;
  
  /* allocate and init structure to assign rooms */  
  for (i=0; i<DAY; i++) {
    reservations[i] = (int *)malloc(roomNum * sizeof(int));
    if (reservations[i] == NULL) {
      fprintf(stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    for (j=0; j<roomNum; j++) {
      reservations[i][j] = -1;
    }
  }

  /* for each request ... */
  for (k=0; k<n; k++) {
    in = requests[k].in - 1;
    out = requests[k].out - 1;

    /* ... reserve all days ... */
    for (i=in; i<out; i++) {
      /* ... in the first available room */
      for (stop=0, j=0; j<roomNum && stop==0; j++) {
        if (reservations[i][j] == -1) {
          reservations[i][j] = k;
          stop = 1;
        }
      }
    }
  }
}

/*
 *  write the output file
 */
void file_write (int *reservations[], int roomNum, request_t *requests) {
  char name[MAX];
  int i, j, k;
  FILE *fp;

  fprintf(stdout, "Output file name: ");
  scanf("%s", name);
  fp = fopen (name, "w");
  if (fp == NULL) {
    fprintf(stderr, "File open error (file=%s).\n", name);
    exit(EXIT_FAILURE);
  } 
  
  fprintf(fp, "Rooms needed = %d\n", roomNum);
  for (j=0; j<roomNum; j++) {
    fprintf(fp, "Room n. %d\n", j+1);
    for (i=0; i<DAY; i++) {
      if (reservations[i][j] != -1) {
        /* Get data from the original structure */
        k = reservations[i][j];
        fprintf(fp, "day=%d %s %s\n", i+1, requests[k].surname, requests[k].name);
      }
    }
  }
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
