/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 03 - exercise 13 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define MAX   100
#define WORD 25+1
#define DAY   365

typedef struct reservation_s {
  char hospital[WORD];
  char unit[WORD];
  int n;
  int **beds;
} reservation_t;

void file_read (int *, reservation_t **);
void menu (int, reservation_t *);
void delete (int, reservation_t *);
void reserve (int, reservation_t *);

int main () {
  reservation_t *reservation;
  int i, j, n;

  file_read (&n, &reservation);
  menu (n, reservation);
  
  for (i=0; i<n; i++) {
    for (j=0; j<reservation[i].n; j++) {
      free (reservation[i].beds[j]);
    }
    free (reservation[i].beds);
  }
  free (reservation);

  return (0);
}

void file_read (int *n, reservation_t **reservationP) {
  FILE *fp;
  reservation_t *reservation;
  char fileName[MAX];
  int i, j, k;

  fprintf (stdout, "Input file name: ");
  scanf ("%s", fileName);

  fp = fopen (fileName, "r");
  if (fp==NULL) {
    fprintf (stderr, "Error opening file!\n");
    exit(EXIT_FAILURE);
  } 
  
  if (fscanf (fp, "%d", n) != 1) {
    fprintf (stderr, "Error reading file!\n");
    exit(EXIT_FAILURE);    
  }
  
  reservation = (reservation_t *) malloc (*n * sizeof (reservation_t));
  if (reservation == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  i = 0;
  while (i<*n &&
    fscanf (fp, "%s%s%d", reservation[i].hospital, reservation[i].unit,
      &reservation[i].n) != EOF) {

    reservation[i].beds = (int **) malloc (reservation[i].n * sizeof (int *));
    if (reservation[i].beds == NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    for (j=0; j<reservation[i].n; j++) {
      reservation[i].beds[j] = (int *) malloc (DAY * sizeof (int));
      if (reservation[i].beds[j] == NULL) {
        fprintf (stderr, "Memory allocation error.\n");
        exit(EXIT_FAILURE);
      }
    }

    /* Init structure */
    for (j=0; j<reservation[i].n; j++) {
      for (k=0; k<DAY; k++) {
        reservation[i].beds[j][k]= 0;
      }
    }

    i++;
  }

  if (*n != i) {
    fprintf (stderr, "Wrong file format!\n");
    *n = i;
  }
  
  *reservationP = reservation;
  fclose (fp);
  
  return;
}

void menu (int n, reservation_t *reservation) {
  char command[MAX];

  do {
    fprintf (stdout, "Comand> ");
    scanf ("%s", command);

    if (strcmp (command, "reserve") == 0) {
      reserve (n, reservation);
    } else
    if (strcmp (command, "remove") == 0) {
      delete (n, reservation);
    } else
    if (strcmp (command, "end") != 0) {
      fprintf (stderr, "Command error!\n");
    }
  } while (strcmp (command, "end") != 0);

  return;
}

void reserve (int n, reservation_t *reservation) {
  char unit[MAX];
  int i, j, k, length, from, to;
  int stop1, stop2;
  int minH, minB, minData;

  scanf ("%s%d", unit, &length);

  minH = minB = minData = (-1);
  /* For each hospital i */
  for (i=0; i<n; i++) {
    /* IFF the unit is the right one */
    if (strcmp (reservation[i].unit, unit) == 0) {
      /* Look for bed available sooner */
      for (j=0; j<reservation[i].n; j++) {
        for (stop1=0, from=0; from<=DAY-length && stop1==0; from++) {
	  to = from + length;
          for (stop2=0, k=from; k<to && stop2==0; k++) {
            if (reservation[i].beds[j][k] != 0) {
              stop2 = 1;
            }
          }
          if (stop2==0) {
            stop1 = 1;
            if ((minData==(-1) || from<minData)) {
              minH = i;
              minB = j;
              minData = from;
            }
          }
        }
      }
    }
  }

  if (minData == (-1)) {
    fprintf (stdout, "Reserve NOT done.\n");
  } else {
    /* Make the reservation */
    for (k=minData; k<(minData+length); k++) {
      reservation[minH].beds[minB][k] = 1;
    }
    fprintf (stdout, "Reserve done: %s %d %d letto %d.\n",
      reservation[minH].hospital, minData+1, minData+length, minB+1);
  }

  return;
}

void delete (int n, reservation_t *reservation) {
  char hospital[MAX], unit[MAX];
  int i, j, k, from, to;
  int stop1, stop2, stop3;

  //fprintf (stdout, "Hospital Unit From To: ");
  scanf ("%s%s%d%d", hospital, unit, &from, &to);

  /* For each hospital */
  for (stop1=0, i=0; i<n && stop1==0; i++) {
    /* IFF the hospital and the unit are the right ones */
    if (strcmp (reservation[i].hospital, hospital) == 0 &&
      strcmp (reservation[i].unit, unit) == 0) {
      /* Look for a bed */
      for (stop2=0, j=0; j<reservation[i].n && stop2==0; j++) {
        for (stop3=0, k=from-1; k<to && stop3==0; k++) {
          if (reservation[i].beds[j][k] != 1) {
            stop3 = 1;
          }
        }
        if (stop3 == 0) {
          stop2 = 1;
          /* Clean the reservation */
          for (k=from-1; k<to; k++) {
            reservation[i].beds[j][k] = 0;
          }
        }
      }

      stop1 = 1;
    }
  }

  if (stop3 == 0) {
    fprintf (stdout, "Delete done.\n");
  } else {
    fprintf (stdout, "Delete NOT done.\n");
  }

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
