/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 03 - exercise 16 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_LINE  1000
#define MAX_WORD   100+1

/* structure declaration */
typedef struct user_s {
  char *name;
  char *telephone;
  char *job;
  char *code;
  int *borrows;
  int num_borrow;
} user_t;

typedef struct book_s {
  char *author;
  char *title;
  char *position;
  char *code;
  int num_copies;
  int *borrows;
  int num_borrow;
} book_t;

typedef struct borrow_s {
  int user;
  int book;
  char *date;
} borrow_t;

/* function prototypes */
user_t *load_users(int *);
book_t *load_books(int *);
borrow_t *load_borrows(user_t *, int, book_t *, int, int *);
int find_user(user_t *, int, char *);
int find_book(book_t *, int, char *);
void print_users(user_t *, int, borrow_t *, book_t *);
void print_books(book_t *, int, borrow_t *, user_t *);
void quit_memory(user_t *, int, book_t *, int, borrow_t *, int);

/*
 *  main program
 */
int main(void) {
  int num_user, num_book, num_borrow;
  user_t *users;
  book_t *books;
  borrow_t *borrows;

  /* load the three input files */
  users = load_users(&num_user);
  books = load_books(&num_book);
  borrows = load_borrows(users, num_user, books, num_book, &num_borrow);

  /* print the two output files */
  print_users(users, num_user, borrows, books);
  print_books(books, num_book, borrows, users);

  quit_memory(users, num_user, books, num_book, borrows, num_borrow);
  return EXIT_SUCCESS;
}

/*
 *  load the user file contents
 */
user_t *load_users(int *num_ptr) {
  char line[MAX_LINE], name[MAX_WORD], phone[MAX_WORD],
       job[MAX_WORD], code[MAX_WORD];
  int i, count=0;
  user_t *users;
  FILE *fp;

  /* count the file rows */
  fp = fopen("users.txt", "r");
  while (fgets(line, MAX_LINE, fp) != NULL) {
    count++;
  }
  fclose(fp);

  /* allocate the user array and save the file contents */
  users = (user_t *)malloc(count*sizeof(user_t));
  fp = fopen("users.txt", "r");
  for (i=0; i<count; i++) {
    fscanf(fp, "%s %s %s %s", name, phone, job, code);
    users[i].name = strdup(name);
    users[i].telephone = strdup(phone);
    users[i].job = strdup(job);
    users[i].code = strdup(code);
    users[i].borrows = NULL;
    users[i].num_borrow = 0;
  }
  fclose(fp);

  *num_ptr = count;
  return users;
}

/*
 *  load the book file contents
 */
book_t *load_books(int *num_ptr) {
  char line[MAX_LINE], author[MAX_WORD], title[MAX_WORD];
  char position[MAX_WORD], code[MAX_WORD];
  int i, num, count=0;
  book_t *books;
  FILE *fp;

  /* count the file rows */
  fp = fopen("books.txt", "r");
  while (fgets(line, MAX_LINE, fp) != NULL) {
    count++;
  }
  fclose(fp);

  /* allocate the book array and save the file contents */
  books = (book_t *)malloc(count*sizeof(book_t));
  fp = fopen("books.txt", "r");
  for (i=0; i<count; i++) {
    fscanf(fp, "%s %s %s %s %d", author, title, position, code, &num);
    books[i].author = strdup(author);
    books[i].title = strdup(title);
    books[i].position = strdup(position);
    books[i].code = strdup(code);
    books[i].num_copies = num;
    books[i].borrows = NULL;
    books[i].num_borrow = 0;
  }
  fclose(fp);

  *num_ptr = count;

  return books;
}

/*
 *  load the borrow file 
 */
borrow_t *load_borrows (
  user_t *users, int num_user, 
  book_t *books, int num_book, int *num_ptr
) {
  char line[MAX_LINE], user_code[MAX_WORD], book_code[MAX_WORD], date[MAX_WORD];
  int i, j, count=0;
  borrow_t *borrows;
  FILE *fp;

  /* count the file rows */
  fp = fopen("borrows.txt", "r");
  while (fgets(line, MAX_LINE, fp) != NULL) {
    count++;
  }
  fclose(fp);

  /* allocate the borrow array and save the file contents */
  borrows = (borrow_t *)malloc(count*sizeof(borrow_t));
  fp = fopen("borrows.txt", "r");
  for (i=0; i<count; i++) {
    fscanf(fp, "%s %s %s", user_code, book_code, date);
    borrows[i].date = strdup(date);

    borrows[i].user = find_user(users, num_user, user_code);
    users[borrows[i].user].num_borrow++;

    borrows[i].book = find_book(books, num_book, book_code);
    books[borrows[i].book].num_borrow++;
  }
  fclose(fp);

  /* arrange the "pointers" from books and users to borrows */
  for (i=0; i<num_user; i++) {
    if (users[i].num_borrow > 0) {
      users[i].borrows = (int *)malloc(users[i].num_borrow*sizeof(int));
      users[i].num_borrow = 0;
    }
  }

  for (i=0; i<num_book; i++) {
    if (books[i].num_borrow > 0) {
      books[i].borrows = (int *)malloc(books[i].num_borrow*sizeof(int));
      books[i].num_borrow = 0;
    }
  }

  for (i=0; i<count; i++) {
    j = borrows[i].user;
    users[j].borrows[users[j].num_borrow++] = i;
    j = borrows[i].book;
    books[j].borrows[books[j].num_borrow++] = i;
  }

  *num_ptr = count;
  return borrows;
}

/*
 *  find a user given his code
 */
int find_user (user_t *users, int dim, char *key) {
  int i;

  for (i=0; i<dim; i++) {
    if (strcmp(key, users[i].code) == 0) {
      return i;
    }
  }

  return -1;
}

/*
 *  find a book given its code
 */
int find_book(book_t *books, int dim, char *key) {
  int i;

  for (i=0; i<dim; i++) {
    if (strcmp(key, books[i].code) == 0) {
      return i;
    }
  }

  return -1;
}

/*
 *  print the users-borrows file
 */
void print_users (
  user_t *users, int num_user, borrow_t *borrows, book_t *books
) {
  int i, j, k, b;
  FILE *fp;

  fp = fopen("users_borrows.txt", "w");
  for (i=0; i<num_user; i++) {
    fprintf(fp, "%s %s ", users[i].name, users[i].telephone);
    fprintf(fp, "%s %s ", users[i].job, users[i].code);
    fprintf(fp, "%d\n", users[i].num_borrow);
    for (j=0; j<users[i].num_borrow; j++) {
      k = users[i].borrows[j];  /* borrow index */
      b = borrows[k].book;      /* book index */
      fprintf(fp, "%s %s\n", books[b].code, borrows[k].date);
    }
  }
  fclose(fp);
}

/*
 *  print the books-borrows file
 */
void print_books(
  book_t *books, int num_book, borrow_t *borrows, user_t *users
) { 
  int i, j, k, u;
  FILE *fp;

  fp = fopen("books_borrows.txt", "w");
  for (i=0; i<num_book; i++) {
    fprintf(fp, "%s %s ", books[i].author, books[i].title);
    fprintf(fp, "%s %s ", books[i].position, books[i].code);
    fprintf(fp, "%d %d\n", books[i].num_copies, books[i].num_borrow);
    for (j=0; j<books[i].num_borrow; j++) {
      k = books[i].borrows[j];  /* borrow index */
      u = borrows[k].user;      /* user index */
      fprintf(fp, "%s %s\n", users[u].code, borrows[k].date);
    }
  }
  fclose(fp);
}

/*
 *  free all allocated memory
 */
void quit_memory(
  user_t *users, int num_user, book_t *books, int num_book,
  borrow_t *borrows, int num_borrow
) {
  int i;

  for (i=0; i<num_user; i++) {
    free(users[i].name);
    free(users[i].telephone);
    free(users[i].job);
    free(users[i].code);
    free(users[i].borrows);
  }
  free(users);

  for (i=0; i<num_book; i++) {
    free(books[i].author); 
    free(books[i].title); 
    free(books[i].position); 
    free(books[i].code); 
    free(books[i].borrows);
  }
  free(books);

  for (i=0; i<num_borrow; i++) {
    free(borrows[i].date);
  }
  free(borrows);
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
