/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 04 - exercise 02 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#define LINE 101

/* structure declaration */
typedef struct word_s {
  char *word;
  int count;
  struct word_s *next;
} word_t;

/* function prototypes */
word_t *input_read (FILE *);
word_t *word_find (word_t *, char *);
word_t *word_sort (word_t *);
void output_write (word_t *, FILE *);
void cleanUp (word_t *);

/* 
 *  main program
 */
int main (int argc, char *argv[]) {
  word_t *head=NULL;
  FILE *fp;

  fp = fopen(argv[1], "r");
  head = input_read(fp);
  fclose(fp);

  head = word_sort(head);

  fp = fopen(argv[2], "w");
  output_write(head, fp);
  fclose(fp);

  cleanUp(head);
  return 0;
}

/* 
 *  count the occurrences of every word in the input file
 */
word_t *input_read (FILE *fp) {
  word_t *tmp, *head=NULL;
  char buffer[LINE];
  int i;

  while (fscanf(fp, "%s", buffer) != EOF) {
    for (i=0; i<strlen(buffer); i++) {
      buffer[i] = tolower(buffer[i]);
    }
    tmp = word_find(head, buffer);
    if (tmp != NULL) {
      tmp->count++;
    } else {
      tmp = (word_t *)malloc(sizeof(word_t));
      tmp->word = strdup(buffer);
      tmp->count = 1;
      tmp->next = head;
      head = tmp;
    }
  }

  return head;
}

/* 
 *  search for a given word in the list of words
 */
word_t *word_find (word_t *head, char *word) {
  word_t *tmp=head;

  while (tmp != NULL) {
    if (strcmp(tmp->word, word) == 0) {
      return tmp;
    }
    tmp = tmp->next;
  }
  return NULL;
}

/* 
 *  sort the list of words (alphabetical order)
 */
word_t *word_sort (word_t *head) {
  word_t *tmp, *scan, *ptr, *sort=NULL;

  while (head != NULL) {
    tmp = head;
    head = head->next;
    if (sort==NULL || sort->count>tmp->count || 
       (sort->count==tmp->count && strcmp(sort->word, tmp->word)>0)) {
      tmp->next = sort;
      sort = tmp;
    } else {
      scan = sort;
      ptr = scan->next;
      while (ptr!=NULL && (ptr->count<tmp->count || 
        (ptr->count==tmp->count && strcmp(ptr->word, tmp->word)<0))) {
        scan = ptr;
        ptr = scan->next;
      }
      scan->next = tmp;
      tmp->next = ptr;
    }
  }
  return sort;
}

/* 
 *  write the output file
 */
void output_write (word_t *head, FILE *fp) {
  word_t *tmp=head;

  while (tmp != NULL) {
    fprintf(fp, "%s %d\n", tmp->word, tmp->count);
    tmp = tmp->next;
  }

  return;
}

/* 
 *  quit the dynamically allocated memory for the list of words
 */
void cleanUp (word_t *head) {
  word_t *tmp;

  while (head != NULL) {
    tmp = head;
    head = head->next;
    free(tmp->word);
    free(tmp);
  }

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
