/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 04 - exercise 03 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define MAX 50+1

/* structure declaration */
typedef struct employee_s {
  char *nc;
  char cf[17];
  int ns;
  int **mat;
  int to;
  struct employee_s *next;     
} employee_t;

/* function prototypes */
employee_t *employee_read (FILE *);
employee_t *list_insert (employee_t *, employee_t *);
void file_write (FILE *, employee_t *);
void memory_free (employee_t *); 

/* 
 *  main program
 */
int main (int argc, char *argv[]) {
  employee_t *e, *head = NULL;
  FILE *fin, *fout;
 
  fin = fopen(argv[1], "r");
  do {
    e = employee_read(fin);
    if (e != NULL) {
      head = list_insert(head, e);
    }
  } while (e != NULL);
  fclose(fin);

  fout = fopen(argv[2], "w");
  file_write(fout, head);
  fclose(fout);
   
  memory_free(head);
  return EXIT_SUCCESS;
}     

/* 
 *  read info for a new employee from file
 */
employee_t *employee_read (FILE *fin) {
  char n[MAX], c[MAX], cf[17], tmp[MAX*2];;
  employee_t *e;
  int i, j, ns, num, day;

  if (fscanf(fin,"%s %s %s %d", c, n, cf, &ns) == EOF) {
    return NULL;
  }

  e = (employee_t *)malloc(sizeof(employee_t));
  sprintf(tmp, "%s %s", c, n);
  e->nc = strdup(tmp);
  strcpy(e->cf, cf); 
  e->ns = ns;
  e->to = 0;

  e->mat = (int **)malloc(ns * sizeof(int *));
  for (i=0; i<ns; i++) {
    fscanf(fin, "%d %d", &num, &day);
    e->mat[i] = (int *)malloc((day+2) * sizeof(int));
    e->mat[i][0] = num;
    e->mat[i][1] = day;
    for (j=0; j<day; j++) {
      fscanf(fin, "%d", &e->mat[i][j+2]);
      e->to = e->to + e->mat[i][j+2];
    }                                
  }

  return e;
}

/* 
 *  add a new element to the list of employees 
 */
employee_t *list_insert (employee_t *head, employee_t *e) {
  employee_t *p;

  if (head==NULL || e->to<head->to) {
    e->next = head;
    return e;
  }

  p = head;  
  while (p->next!=NULL && p->next->to<e->to) {
    p = p->next;
  } 
  e->next = p->next;                    
  p->next = e;

  return head;
}

/* 
 *  write the output file
 */
void file_write (FILE *fp, employee_t *head) {
  employee_t *e=head;
  int i, j;
    
  while (e != NULL){
    fprintf(fp, "%s %s %d\n", e->nc, e->cf, e->ns);
    for (i=0; i<e->ns; i++) {
      for (j=0; j<e->mat[i][1]+2; j++) {                              
        fprintf(fp, "%d ", e->mat[i][j]);
      }
      fprintf(fp, "\n");
    }  
    e = e->next;
  }
}

/* 
 *  free all the dynamically allocated memory
 */
void memory_free (employee_t *head) {
  employee_t *e;
  int i;

  while (head != NULL) {
    for (i=0; i<head->ns; i++){
      free(head->mat[i]);
    }     
    free(head->mat);
    free(head->nc);
    e = head;
    head = head->next;
    free(e);
  }
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
