/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 04 - exercise 08 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MEM 8000

/* structure declaration */
typedef struct memory_s {
  int ptr, size;
  struct memory_s *next;
} memory_t;

/* function prototypes */
int myAlloc(memory_t *, int);
int myFree(memory_t *, int);
void display(memory_t *);
void quit(memory_t *);

/*
 *  main program
 */
int main (void) {
  memory_t *pTop=NULL, *pNew;
  int size, ptr, val;
  char cmd[20];

  pNew = (memory_t *)malloc(sizeof(memory_t));
  if (pNew==NULL){
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  pNew->size = 0;
  pNew->ptr = MEM;
  pNew->next = NULL;
  pTop = (memory_t *)malloc(sizeof(memory_t));
  if (pTop==NULL){
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  pTop->size = 0;
  pTop->ptr = 0;
  pTop->next = pNew;

  do {
    fprintf(stdout, "Input command: ");
    scanf("%s", cmd);
    if (strcmp(cmd, "malloc")==0 || (strcmp(cmd, "free") == 0)) {
      scanf("%d", &val);
      if (strcmp(cmd, "malloc") == 0) {
        ptr = myAlloc(pTop, val);
        if (ptr >= 0) {
          fprintf(stdout, "Allocation (size=%d) returns position %d\n", val, ptr);
        } else {
          fprintf(stdout, "Memory allocation was not possible!\n");
        }
      } else {
        size = myFree(pTop, val);
        if (size > 0) {
          fprintf(stderr, "Memory block with size %d freed\n", size);
        } else {
          fprintf(stderr, "Memory block not found\n");
        }
      }
      display(pTop);
    } else if (strcmp(cmd, "stop") != 0) {
      fprintf(stderr, "Error: unknown command\n");
    }
  } while (strcmp(cmd, "stop") != 0);

  quit(pTop);
  return EXIT_SUCCESS;
}

/*
 *  "allocate" a new memory block, if possible
 */
int myAlloc (memory_t *pTop, int size) {
  memory_t *pNew, *tmp=pTop;
  int mem;

  while (tmp->next != NULL) {
    mem = tmp->next->ptr - (tmp->ptr+tmp->size);
    if (size <= mem) {
      pNew = (memory_t *)malloc(sizeof(memory_t));
      if (pNew==NULL){
        fprintf(stderr, "Memory allocation error.\n");
        exit(EXIT_FAILURE);
      }
      pNew->size = size;
      pNew->ptr = tmp->ptr + tmp->size;
      pNew->next = tmp->next;
      tmp->next = pNew;
      return pNew->ptr;
    }
    tmp = tmp->next;
  }
  return -1;
}

/*
 *  "free" a memory block 
 */
int myFree (memory_t *pTop, int ptr) {
  memory_t *pOld, *tmp=pTop;
  int size;

  while (tmp->next != NULL) {
    if (tmp->next->ptr == ptr) {
      pOld = tmp->next;
      tmp->next = pOld->next;
      size = pOld->size;
      free(pOld);
      return size;
    }
    tmp = tmp->next;
  }
  return -1;
}

/*
 *  show the memory usage
 */
void display (memory_t *pTop) {
  memory_t *tmp=pTop->next;

  fprintf(stdout, "Memory usage:\n");
  while (tmp->ptr != MEM) {
    fprintf(stdout, "\t%d (%d)\n", tmp->ptr, tmp->size);                  
    tmp = tmp->next;
  }
  fprintf(stdout, "\n");

  return;
}

/*
 *  free the dynamically allocated memory
 */
void quit (memory_t *pTop) {
  memory_t *tmp;

  while (pTop != NULL) {
    tmp = pTop;
    pTop = pTop->next;
    free(tmp);
  }

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
