/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 04 - exercise 10 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#define LENGTH 100

/* structure declaration */
typedef struct segment_s {
  float x1;
  float y1;
  float x2;
  float y2;
  struct segment_s *next;
} segment_t;

/* function prototypes */
segment_t *read(char *name);
int check(segment_t *head);
int sticky(segment_t *ptr1, segment_t *ptr2, int check_reverse);
void swap(segment_t *ptr1, segment_t *ptr2);
void write(char *name, segment_t *head);

/* 
 *  main program
 */
int main (void) {
  segment_t *head;
  char name[LENGTH];
  int found;

  fprintf(stdout, "Input file name: ");
  scanf("%s", name);
  head = read(name);

  found = check(head);

  fprintf(stdout, "Output file name: ");
  scanf("%s", name);
  if (found)  {
    write(name, head);
  } else {
    write(name, NULL);
  }

  return EXIT_SUCCESS;
}

/* 
 *  read the input file contents
 */
segment_t *read (char *name) {
  FILE *fp;
  segment_t *head=NULL, *tmp;
  float x1, y1, x2, y2;

  fp = fopen(name, "r");
  head = NULL;
  while (fscanf(fp, "%f %f %f %f", &x1, &y1, &x2, &y2) != EOF) {
    tmp = (segment_t *)malloc(sizeof(segment_t));
    if (tmp==NULL){
      fprintf(stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    tmp->x1 = x1;
    tmp->y1 = y1;
    tmp->x2 = x2;
    tmp->y2 = y2;
    tmp->next = head;
    head = tmp;
  }
  fclose (fp);

  return head;
}

/* 
 *  check whether the set of segments define a polygon
 */
int check (segment_t *head) {
  segment_t *tmp1, *tmp2, *ptr;
  int stop, found;

  stop = 0;
  ptr = NULL;
  tmp1 = head;
  while (tmp1!=NULL && stop==0) {
    stop = 1;

    /* look for the next segment */
    found = 0;
    tmp2 = tmp1->next;
    while (tmp2!=NULL && found==0) {
      if (sticky(tmp1, tmp2, 1)) {
	found = 1;

	/* tie current segment to the previous one */
	swap(tmp1->next, tmp2);
	stop = 0;
      }
      tmp2 = tmp2->next;
    }
    
    ptr = tmp1;
    tmp1 = tmp1->next; 
  }

  /* check the reason for which the loop stopped ... */
  if (tmp1 != NULL) {
    /* not all segments have been tied: poygon NOT found */
    return 0;
  } 

  /* all segments tied: is last segment also sticking to first one? */
  return sticky(ptr, head, 0);
}

/* 
 *  check whether two segments have a common extreme
 */
int sticky (segment_t *ptr1, segment_t *ptr2, int check_reverse) {
  float x, y;

  if (ptr1->x2==ptr2->x1 && ptr1->y2==ptr2->y1) {
    return 1;
  }

  if (check_reverse && ptr1->x2==ptr2->x2 && ptr1->y2==ptr2->y2) {
    /* swap the two extremes */
    x = ptr2->x1;
    y = ptr2->y1;
    ptr2->x1 = ptr2->x2;
    ptr2->y1 = ptr2->y2;
    ptr2->x2 = x;
    ptr2->y2 = y;

    return 1;
  }

  return 0;
}

/* 
 *  swap two elements of the list
 */
void swap (segment_t *ptr1, segment_t *ptr2) {
  segment_t tmp;

  tmp = *ptr1;
  *ptr1 = *ptr2;
  ptr1->next = tmp.next;

  tmp.next = ptr2->next;
  *ptr2 = tmp;

  return;
}

/* 
 *  write the output file
 */
void write (char *name, segment_t *head) {
  FILE *fp;
  segment_t *tmp;

  fp = fopen(name, "w");
  if (fp==NULL) {
    fprintf (stderr, "Error opening file!\n");
    exit(EXIT_FAILURE);
  }
 
  if (head != NULL) {
    fprintf(fp, "Polygon found\n");
    for (tmp=head; tmp!=NULL; tmp=tmp->next) {
      fprintf(fp, "%.1f %.1f %.1f %.1f\n", tmp->x1, tmp->y1, tmp->x2, tmp->y2);
    }
  } else {
    fprintf(fp, "Polygon NOT found\n");
  }

  fclose(fp);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
