/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 04 - exercise 11 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX  101

/* structure declaration */
typedef struct trait trait_t;
typedef struct member member_t;

struct trait {
  char *id;
  trait_t *next;
};

struct member {
  char *name;
  char sex;
  trait_t *features;
  member_t *next;
};

/* function prototypes */
member_t *load(char *filename);
trait_t *read(char *sex_ptr);
void process(member_t *club, char sex, trait_t *features);
int check(trait_t *head1, trait_t *head2);
void destroy(trait_t *head);
void quit(member_t *head);

/*
 *  main program
 */
int main (void) {
  member_t *club=NULL;
  trait_t *features=NULL;
  char name[MAX], sex;

  fprintf(stdout, "Input the club file name: ");
  scanf("%s", name);
  club = load(name);

  do {
    features = read(&sex);
    if (features) {
      process(club, sex, features);
      destroy(features);
    }
  } while (features != NULL);

  quit(club);
  
  return EXIT_SUCCESS;
}

/*
 *  load the info for the club members
 */
member_t *load (char *filename) {
  char lastName[MAX], firstName[MAX], line[MAX], name[MAX], sex[MAX];
  member_t *member, *head=NULL;
  trait_t *feature;
  FILE *fp;

  fp = fopen(filename, "r");
  while (fgets(line, MAX, fp) != NULL) {
    if (line[0] == '#') {
      member = (member_t *)malloc(sizeof(member_t));
      if (member==NULL){
        fprintf(stderr, "Memory allocation error.\n");
        exit(EXIT_FAILURE);
      }
      sscanf(&line[1], "%s %s %s", lastName, firstName, sex);
      sprintf(name, "%s %s", lastName, firstName);
      member->name = strdup(name);
      member->features = NULL;
      member->sex = sex[0];
      member->next = head;
      head = member;
    } else {
      feature = (trait_t *)malloc(sizeof(trait_t));
      if (feature==NULL){
        fprintf(stderr, "Memory allocation error.\n");
        exit(EXIT_FAILURE);
      }
      sscanf(line, "%s %s", lastName, firstName);
      sprintf(name, "%s %s", lastName, firstName);
      feature->id = strdup(name);
      feature->next = member->features;
      member->features = feature;
    }
  }
  fclose(fp);

  return head;
}

/*
 *  read a user's query (list of features)
 */
trait_t *read (char *sex_ptr) {
  char line[MAX], name[MAX], value[MAX], id[MAX], sex, *ptr;;
  trait_t *trait, *features=NULL;
  int n;

  fprintf(stdout, "Input query: ");
  fgets(line, MAX, stdin);
  sscanf(line, "%c", &sex);

  if (sex == 'T') {
    /* termination */
    return NULL;
  }

  ptr = &line[1];
  // %n writes into variable n the number of characters that have been read
  while (sscanf(ptr, "%s %s%n", name, value, &n) != EOF) {
    trait = (trait_t *)malloc(sizeof(trait_t));
    if (trait==NULL){
      fprintf(stdout, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    sprintf(id, "%s %s", name, value);
    trait->id = strdup(id);
    trait->next = features;
    features = trait;
    ptr += n;
  } 
  *sex_ptr = sex;
  
  return features;
}

/*
 *  search the club member having most traits in the given set 
 */
void process (member_t *club, char sex, trait_t *features) {
  member_t *mate=NULL, *ptr=club;
  int equal, max=0;

  while (ptr != NULL) {
    if (ptr->sex == sex) {
      equal = check(ptr->features, features);
      if (mate==NULL || equal>max) {
        mate = ptr;
        max = equal;
      }
    }
    ptr = ptr->next;
  }
  fprintf(stdout, "Member found: %s (score %d)\n", mate->name, max);

  return;
}

/*
 *  counts the number of common features
 */
int check (trait_t *head1, trait_t *head2) {
  trait_t *ptr1, *ptr2;
  int found, equal=0;

  ptr1 = head1;
  while (ptr1 != NULL) {
    ptr2 = head2;
    found = 0;
    while (ptr2!=NULL && !found) {
      if (strcmp(ptr1->id, ptr2->id) == 0) {
        found = 1;
      }
      ptr2 = ptr2->next;
    }
    equal += found;
    ptr1 = ptr1->next;
  }
  return equal;
}

/*
 *  free the memory allocated for a trait list
 */
void destroy (trait_t *head) {
  trait_t *tmp;

  while (head != NULL) {
    tmp = head;
    head = tmp->next;
    free(tmp->id);
    free(tmp);
  }
}

/*
 *  free the memory allocated for the entire club
 */
void quit (member_t *head) {
  member_t *tmp;

  while (head != NULL) {
    tmp = head;
    head = tmp->next;
    free(tmp->name);
    destroy(tmp->features);
    free(tmp);
  }
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
