/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 05 - exercise 07 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#define DIM 20

/* function prototypes */
void map_read (char *, int [DIM][DIM], int *, int *);
void expand_r (int [DIM][DIM], int, int, int);
void map_write (char *, int [DIM][DIM], int, int, int);

/*
 * main program
 */
int main (int argc, char *argv[]) {
  int map[DIM][DIM];
  int i, j, nr, nc, n=0;

  map_read(argv[1], map, &nr, &nc);
  for (i=0; i<nr; i++) {
    for (j=0; j<nc; j++) {
      if (map[i][j] == (-1)) {
        expand_r(map, i, j, ++n);
      }
    }
  }
  map_write(argv[2], map, nr, nc, n);

  return EXIT_SUCCESS;
}

/*
 *  load the map from file
 */
void map_read (char *name, int map[DIM][DIM], int *nr, int *nc) {
  FILE *fp;
  char line[DIM+2];
  int i, j;

  fp = fopen(name, "r");

  i = 0;
  while (fgets(line, DIM+2, fp) != NULL) {
    j = 0;
    while (j<DIM && (line[j]=='.' || line[j]=='*')) {
      map[i][j] = (line[j]=='*' ? 0 : -1);
      j++;
    }
    i++;
  }

  fclose(fp);

  *nr = i;
  *nc = j;
  return;
}

/*
 *  expand a region numbering it, recursive function
 */
void expand_r (int map[DIM][DIM], int x, int y, int id) {
  int i, xx, yy;
  const int xoff[] = {1, 0, -1, 0};
  const int yoff[] = {0, 1, 0, -1};

  if (map[x][y] != (-1)) {
    return;
  }

  /* mark the position */
  map[x][y] = id;

  /* look for adjacences */
  for (i=0; i<4; i++) {
    xx = x + xoff[i];
    yy = y + yoff[i];
    expand_r(map, xx, yy, id);
  }
}

/*
 *  write map on outptu file
 */
void map_write (char *name, int map[DIM][DIM], int nr, int nc, int n) {
  FILE *fp;
  int i, j;

  fp = fopen(name, "w");

  for (i=0; i<nr; i++) {
    for (j=0; j<nc; j++) {
      if (map[i][j]==0) {
        fprintf (fp, "*");
      } else {
        fprintf (fp, "%d", map[i][j]);
      }
    }
    fprintf (fp, "\n");
  }
  fprintf (stdout, "Regions number = %d\n", n);

  fclose (fp);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
