/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 05 - exercise 08 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* function prototypes */
float eval_r (char *expr, int *pos_ptr);

/*
 *  main program
 */
int main (int argc, char *argv[]) {
  float result;
  int pos=0;

  if (argc < 2) {
     fprintf(stderr, "Error: missing parameter.\n");
     fprintf(stderr, "Run as: %s \"<prefix_expression>\"\n", argv[0]);
     return 1;
  }

  result = eval_r(argv[1], &pos);
  fprintf(stdout, "Result = %.2f\n", result);
  return EXIT_SUCCESS;
}

/*
 *  evaluate a prefix expression, recursive function
 */
float eval_r (char *expr, int *pos_ptr) {
  float left, right, result;
  char operator;
  int k = *pos_ptr;

  /* skip spaces */
  while (isspace(expr[k])) {
    k++;
  }

  /* manage sub-expressions and compute the result */
  if (expr[k]=='+' || expr[k]=='*' || expr[k]=='-' || expr[k]=='/') {
    operator = expr[k++];
    left = eval_r(expr, &k);
    right = eval_r(expr, &k);
    switch (operator) {
      case '+': result = left+right; break;
      case '*': result = left*right; break;
      case '-': result = left-right; break;
      case '/': result = left/right; break;
    }
  } else {
    /* terminal case: just a real value */
    sscanf(&expr[k], "%f", &result);
    while (isdigit(expr[k]) || expr[k]=='.') {
      k++;
    }
  }

  *pos_ptr = k;

  return result;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
