/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 06 - exercise 03 - solution 0
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

/*
 * Select (1) One (and only one) Partition Scheme:
 * HOARE Version 1 with pivot = rigthmost element
 * HOARE Version 1 with pivot = leftmost element
 * LOMUTO
 */
#define HOARE_V1  1
#define HOARE_V2  0
#define LOMUTO    0

/* function prototypes */
int *array_read (int *);
void quick_sort_r (int *, int, int);
int partition (int *, int, int);
void swap (int *, int, int);
void array_write(int *, int, int);

/*
 *  main program
 */
int main (void) {
  int n, *v;

  v = array_read(&n);
  quick_sort_r(v, 0, n-1);
  array_write(v, 0, n-1);
  free(v);

  return EXIT_SUCCESS;
}

/*
 *  load the initial array
 */
int *array_read (int *dim) {
  int i, *v;

  fprintf(stdout, "Size: ");
  scanf("%d", dim);
  v = (int *)malloc((*dim)*sizeof(int));
  if (v==NULL){
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  fprintf(stdout, "Initial array:\n");
  for (i=0; i<*dim; i++) {
    fprintf(stdout, "v[%d] = ", i);
    scanf("%d", &v[i]);
  }
  return v;
}

/*
 *  sort an array through quick sort, recursive function
 */
void quick_sort_r(int *v, int l, int r) {
  int q;

  if (l < r) {
    q = partition(v, l, r);
    quick_sort_r(v, l, q-1);
    quick_sort_r(v, q+1, r);
  }
}

/*
 * Lomuto Partition Scheme
 */
#if LOMUTO
int partition(int *v, int l, int r) {
  int i, j, pivot;

  pivot = v[r];
  i = l;
  for (j=l; j<=r-1; j++) {
    if (v[j] <= pivot) {
      swap (v, i, j);
      i++;
    }
  }
  swap (v, i, r);

  return i;
}
#endif

/*
 *  Hoare Partition Scheme:
 *  Pivot = Rightmost element
 */
#if HOARE_V1
int partition(int *v, int l, int r) {
  int i, j, pivot;

  i = l-1;
  j = r;
  pivot = v[r];
  while (i < j) {
    while (v[++i] < pivot) ;
    while (j > l && v[--j] >= pivot) ;

    if (i < j) {
      swap (v, i, j);
    }
  }

  swap (v, i, r);

  return i;
}
#endif

/*
 *  Hoare Partition Scheme:
 *  Pivot = Leftmost element
 */
#if HOARE_V2
int partition (int *v, int l, int r) {
  int i, j, pivot;

  i = l;
  j = r + 1;
  pivot = v[l];
  while (i < j) {
    while (i < r && v[++i] <= pivot);
    while (v[--j] > pivot);

    if (i < j) {
      swap (v, i, j);
    }
  }

  swap (v, l, j);

  return j;
}
#endif

/*
 *  swap
 */
void swap (int *v, int i, int j) {
  int tmp;

  tmp = v[i];
  v[i] = v[j];
  v[j] = tmp;

  return;
}

/*
 *  print the output array
 */
void array_write (int v[], int l, int r) {
  int i;

  fprintf(stdout, "Sorted array:\n");
  for (i=l; i<=r; i++) {
    fprintf(stdout, "v[%d] = %d\n", i, v[i]);
  }
  fprintf(stdout, "\n");
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
