/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 06 - exercise 11 - solution 4
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#define N 8
#define FULL  (+1)
#define EMPTY (-1)

/* function prototypes */
int place_r (int, int *, char *, char *, char *);
void set_queen (int, int, int *, char *, char *, char *);
void rem_queen (int, int, int *, char *, char *, char *);

/*
 *  main program
 */
int main (void) {
  char col[N], diag[2*N-1], revdiag[2*N-1];
  int i, j, queen[N];

  // Reset board (no queen at the beginning)
  for (i=0; i<N; i++) {
    queen[i] = col[i] = EMPTY;
  }
  for (i=0; i<2*N-1; i++) {
    diag[i] = revdiag[i] = EMPTY;
  }

  if (place_r(0, queen, col, diag, revdiag)) {
    fprintf(stdout, "Solution found:\n");
    for(i=0; i<N; i++) {
      for(j=0; j<N; j++) {
        if (queen[i] == j) {
          fprintf(stdout, "Q");
        } else {
          fprintf(stdout, ".");
        }
      }
      fprintf(stdout, "\n");
    }
  } else {
    fprintf(stdout, "Solution NOT found!\n");
  }

  return EXIT_SUCCESS;
}

/*
 *  place N queens on the NxN board, recursive function
 */
int place_r (int n, int *queen, char *col, char *diag, char *revdiag) {
  int r, c;

  // When n==N all queens have been placed on the board
  if (n == N) {
    return 1;
  }

  // Queen n is on row n
  r = n;
  // On which column (c) do I place queen n ?
  for (c=0; c<N; c++) {
    if (col[c]==EMPTY && diag[r+c]==EMPTY && revdiag[r-c+N-1]==EMPTY) {
      set_queen(n, c, queen, col, diag, revdiag);
      if (place_r(n+1, queen, col, diag, revdiag)) {
        return 1;
      }
      rem_queen(r, c, queen, col, diag, revdiag);
    }
  }

  return 0;
}

/*
 *  mark the reachable "cells" when a queen is placed in position i,j
 */
void set_queen (int i, int j, int *queen, char *col, char *diag, char *revdiag) {
  queen[i] = j;
  col[j] = FULL;
  diag[i+j] = FULL;
  revdiag[i-j+N-1] = FULL;
  
  return;
}

/*
 *  remove a queen from position i,j
 */
void rem_queen (int i, int j, int *queen, char *col, char *diag, char *revdiag) {
  queen[i] = EMPTY;
  col[j] = EMPTY;
  diag[i+j] = EMPTY;
  revdiag[i-j+N-1] = EMPTY;

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
