/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 07 - exercise 02 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#define SIZE  10
#define MAX  512

/* function prototypes */
int load (float matrix[SIZE][SIZE]);
float delta_r (float matrix[SIZE][SIZE], int dim);
void reduce (float matrix[SIZE][SIZE], int dim, int col, float rm[SIZE][SIZE]);

/*
 *  main program
 */
int main (void) {
  float matrix[SIZE][SIZE];
  int dim;

  dim = load(matrix);
  fprintf(stdout, "Determinant = %f.\n", delta_r(matrix, dim));
  return EXIT_SUCCESS;
}

/*
 *  read in the matrix file
 */
int load (float matrix[SIZE][SIZE]) {
  char line[MAX], name[20];
  int i, j, n;
  FILE *fp;

  fprintf(stdout, "Input file name: ");
  scanf("%s", name);
  fp = fopen(name, "r");

  /* count the number of rows */
  while (fgets(line, SIZE, fp) != NULL) {
    n++;
  }
  fclose(fp);

  /* read and store the matrix elements */
  fp = fopen(name, "r");
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      fscanf(fp, "%f", &matrix[i][j]);
    }
  }
  fclose(fp);

  return n;
}

/*
 *  compute the determinant of a square matrix, recursive function
 */
float delta_r (float matrix[SIZE][SIZE], int dim) {
  float d=0, smaller[SIZE][SIZE];
  int j, sign=1;

  if (dim == 1) {
    return matrix[0][0];
  }

  for (j=0; j<dim; j++) {
    reduce(matrix, dim, j, smaller);
    d += sign * matrix[0][j] * delta_r(smaller, dim-1);
    sign = -sign;
  }

  return d;
}

/*
 *  reduce a matrix by removing the first row and a specified column
 */
void reduce (
  float matrix[SIZE][SIZE], int dim, int col, float smaller[SIZE][SIZE]
) {
  int i, j, k=0;

  for (i=1; i<dim; i++) {
    for (j=0; j<dim; j++) {
      if (j != col) {
        smaller[k/(dim-1)][k%(dim-1)] = matrix[i][j];
        k++;
      }
    }
  }

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
