/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 07 - exercise 05 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define MAXBUFFER 128

/* function prototypes */
int **load (int *n_ptr);
int find_r (int **schema, int n, int step);
int check (int **sudoku, int n, int step);

/*
 *  main program
 */
int main (void) {
  int **sudoku, dim, i, j;

  sudoku = load(&dim);

  if (find_r(sudoku, dim, 0)) {
    fprintf(stdout, "Solution:\n");
    for (i=0; i<dim; i++) {
      for (j=0; j<dim; j++) {
         fprintf(stdout, "%3d ", sudoku[i][j]);
      }
      fprintf(stdout, "\n");
    }
  } else {
    fprintf(stdout, "No solution found.\n");
  }

  for (i=0; i<dim; i++) {
    free(sudoku[i]);
  }
  free(sudoku);

  return EXIT_SUCCESS;
}

/*
 *  load the initial sudoku scheme from file
 */
int **load (int *dim_ptr) {
  char buffer[MAXBUFFER], filename[20];
  int i, j, **sudoku, dim=0;
  FILE *fp;

  fprintf(stdout, "Input the file name: ");
  scanf("%s", filename);

  fp = fopen(filename, "r");

  /* count the number of rows to get the sudoku size */
  while (fgets(buffer, MAXBUFFER, fp) != NULL) {
    dim++;
  }
  fclose(fp);

  /* allocazione memoria necessaria */
  sudoku = (int **)malloc(dim*sizeof(int *));
  if (sudoku==NULL){
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<dim; i++) {
    sudoku[i] = (int *)malloc(dim*sizeof(int));
    if (sudoku[i]==NULL){
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
  }

  /* store the sudoku matrix */
  fp = fopen(filename, "r");
  for (i=0; i<dim; i++) {
    for (j=0; j<dim; j++) {
      fscanf(fp, "%d", &sudoku[i][j]);
    }
  }
  fclose(fp);

  *dim_ptr = dim;
  return sudoku;
}

/*
 *  recursively assign values to the empty sudoku cells
 */
int find_r (int **sudoku, int dim, int step) {
  int i, j, k;

  if (step >= dim*dim) { 
    /* solution found */
    return 1;
  }

  /* get the current cell position */
  i = step / dim;
  j = step % dim;

  if (sudoku[i][j] != 0) {
    /* originally full cell: no attempt to do */
    return find_r(sudoku, dim, step+1);
  }

  /* try all values from 1 to dim (i.e., n*n) */
  for (k=1; k<=dim; k++) {
    sudoku[i][j] = k;
    if (check(sudoku, dim, step)) {
      if (find_r(sudoku, dim, step+1)) {
        return 1;
      }
    }
    sudoku[i][j] = 0;
  }

  return 0;
}

/*
 *  check whether the game rules are violated
 */
int check (int **sudoku, int dim, int step) {
  int i, j, ii, jj, ib, jb, val, n=floor(sqrt(dim));

  /* get the current cell and block position */
  i = step / dim;
  j = step % dim;
  ib = (i/n) * n;
  jb = (j/n) * n;
  val = sudoku[i][j];

  /* check the row */
  for (jj=0; jj<dim; jj++) {
    if (sudoku[i][jj]==val && jj!=j) {
      return 0;
    }
  }

  /* check the column */
  for (ii=0; ii<dim; ii++) {
    if (sudoku[ii][j]==val && ii!=i) {
      return 0;
    }
  }

  /* check the block */
  for (ii=ib; ii<ib+n; ii++) {
    for (jj=jb; jj<jb+n; jj++) {
      if (sudoku[ii][jj]==val && (ii!=i || jj!=j)) {
        return 0;
      }
    }
  }

  return 1;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
