/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 07 - exercise 07 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#define CODE 10+1
#define NAME 30+1

/* structure declaration */
typedef struct {
  char code[CODE];
  float price;
  int shelves;
  int *heights;
  float *widths;
  float *weights;
} bookcase_t;

typedef struct {
  char name[NAME];
  float height;
  float width;
  float weight;
  int shelf;
} book_t;

/* function prototypes */
bookcase_t *read_models(int *num_ptr);
void sort_models(bookcase_t *models, int left, int right);
book_t *read_material(int *num_ptr);
int suitable_r(bookcase_t *bookcase, book_t *material, int num_books, int idx);

/*
 *  main program
 */
int main (void) {
  int i, j, num_models, num_books, found;
  bookcase_t *models;
  book_t *material;

  /* load info */
  models = read_models(&num_models);
  material = read_material(&num_books);

  /* solve the problem */
  for (i=0; i<num_models && !found; i++) {
    found = suitable_r(&models[i], material, num_books, 0);
    if (found != 0) {
      fprintf(stdout, "Bookcase model to buy: %s.\n", models[i].code);
      fprintf(stdout, "Books disposition:\n");
      for (j=0; j<num_books; j++) {
        fprintf(stdout, "%s --> shelf %d\n",
          material[j].name, material[j].shelf+1);
      }
    }
  }

  /* free memory */
  for (i=0; i<num_models; i++) {
    free(models[i].heights);
    free(models[i].widths);
    free(models[i].weights);
  }
  free(models);
  free(material);

  return EXIT_SUCCESS;
}

/*
 *  read in the bookcases info
 */
bookcase_t *read_models (int *num_ptr) {
  int n, i, j, shelves, width, height, weight;
  bookcase_t *models;
  char name[100];
  float price;
  FILE *fp;

  fprintf(stdout, "Input the bookcases file name: ");
  scanf("%s", name);
  fp = fopen(name, "r");
  fscanf(fp, "%d", &n);
  models = (bookcase_t *)malloc(n*sizeof(bookcase_t));
  if (models==NULL){
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<n; i++) {
    fscanf(fp, "%s %f %d %d %d", name, &price, &shelves, &width, &weight);
    strcpy(models[i].code, name);
    models[i].price = price;
    models[i].shelves = shelves;
    models[i].heights = (int *)malloc(shelves*sizeof(int));
    models[i].widths = (float *)malloc(shelves*sizeof(float));
    models[i].weights = (float *)malloc(shelves*sizeof(float));
    if (models[i].heights==NULL || models[i].widths==NULL ||
	models[i].weights==NULL){
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    for (j=0; j<shelves; j++) {
      fscanf(fp, "%d", &height);
      models[i].widths[j] = width;
      models[i].heights[j] = height;
      models[i].weights[j] = weight;
    }
  }
  fclose(fp);

  sort_models(models, 0, n-1);
  *num_ptr = n;
  return models;
}

/*
 *  sort bookcases by increasing price 
 */
void sort_models (bookcase_t *models, int left, int right) {
  bookcase_t tmp;
  float price;
  int i, j;

  price = models[left].price; 
  i = left-1;
  j = right+1;
  while (i < j) {
    while (models[++i].price < price) ;
    while (models[--j].price > price) ;
    if (i < j) {
      tmp = models[i];
      models[i] = models[j];
      models[j] = tmp;
    }
  }
  if (left < right) {
    sort_models(models, left, j);
    sort_models(models, j+1, right);
  }
}

/*
 *  read in the books info
 */
book_t *read_material (int *num_ptr) {
  float height, width, weight;
  book_t *material;
  char name[100];
  int n, i;
  FILE *fp;

  fprintf(stdout, "Input the books file name: ");
  scanf("%s", name);
  fp = fopen(name, "r");
  fscanf(fp, "%d", &n);
  material = (book_t *)malloc(n*sizeof(book_t));
  if (material==NULL){
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<n; i++) {
    fscanf(fp, "%s %f %f %f", name, &height, &width, &weight);
    strcpy(material[i].name, name);
    material[i].height = height;
    material[i].width = width;
    material[i].weight = weight;
    material[i].shelf = -1;
  }

  fclose(fp);
  *num_ptr = n;
  return material;
}

/*
 *  recursevly check whether a given bookcase can contain all the material
 */
int suitable_r (bookcase_t *bookcase, book_t *material, int num_books, int idx) {
  int i;

  if (idx == num_books) {
    return 1;
  }

  for (i=0; i<bookcase->shelves; i++) {
    if (bookcase->heights[i]>=material[idx].height && 
        bookcase->widths[i]>=material[idx].width &&
        bookcase->weights[i]>=material[idx].weight) {

      bookcase->widths[i] -= material[idx].width;
      bookcase->weights[i] -= material[idx].weight;
      material[idx].shelf = i;
      if (suitable_r(bookcase, material, num_books, idx+1)) {
        return 1;
      }
      bookcase->widths[i] += material[idx].width;
      bookcase->weights[i] += material[idx].weight;
      material[idx].shelf = -1;
    }
  }

  return 0;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
