/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 07 - exercise 09 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

/* structure declaration */
typedef struct {
  int id, n1, n2, taken;
} piece_t;

/* function prototypes */
int chain_r(int *currChain, int currLength,int *bestChain, int bestLength, 
            piece_t *pieces, int num);

/*
 *  main program
 */
int main (int argc, char **argv) {
  int *currChain, *bestChain, currLength, bestLength=0;
  int start, num=0, i, j, k;
  piece_t *pieces;
  char line[20];
  FILE *fp;

  /* count the number of pieces in the input file */
  fp = fopen(argv[1], "r");
  while (fgets(line, 20, fp) != NULL) {
    num++;
  }
  fclose(fp);

  /* allocate arrays */
  pieces = (piece_t *)malloc((num+1)*sizeof(piece_t));
  currChain = (int *)malloc(num*sizeof(int));
  bestChain = (int *)malloc(num*sizeof(int));
  if (pieces==NULL || currChain==NULL || bestChain==NULL){
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  /* store the pieces */
  fp = fopen(argv[1], "r");
  for (i=1; i<=num; i++) {
    fscanf(fp, "%d %d-%d", &pieces[i].id, &pieces[i].n1, &pieces[i].n2);
    pieces[i].taken = 0;
  }
  fclose(fp);

  /* look for the first piece */
  sscanf(argv[2], "%d", &start);
  for (i=1; i<=num && pieces[i].id!=start; i++) ;
  pieces[i].taken = 1;
  currLength = 1;

  /* search the solution: try the first piece "straight" */
  currChain[0] = i;
  bestLength = chain_r(currChain, currLength, bestChain, bestLength, 
                       pieces, num);

  /* search the solution: try the first piece "inverted" */
  currChain[0] = -i;
  bestLength = chain_r(currChain, currLength, bestChain, bestLength, 
                       pieces, num);

  /* output result */
  fp = fopen(argv[3], "w");
  fprintf(fp, "%d\n", bestLength);
  for (i=0; i<bestLength; i++) {
    j = bestChain[i];
    k = abs(j);
    if (j > 0) {
      fprintf(fp, "%d %d-%d\n", pieces[k].id, pieces[k].n1, pieces[k].n2);
    } else {
      fprintf(fp, "%d %d-%d\n", pieces[k].id, pieces[k].n2, pieces[k].n1);
    }
  }
  fclose(fp);

  free(pieces);
  free(bestChain);
  free(currChain);
  return EXIT_SUCCESS;
}

/*
 *  find the longest chain that can be formed
 */
int chain_r (
  int *currChain, int currLength,int *bestChain, int bestLength,
  piece_t *pieces, int num
) {
  int j, lastNumber, stop;

  if (currChain[currLength-1] > 0) {
    lastNumber = pieces[currChain[currLength-1]].n2;
  } else {
    lastNumber = pieces[-currChain[currLength-1]].n1;
  }

  stop = 1;
  for (j=1; j<=num; j++) {
    if (pieces[j].taken == 0) {
      if (pieces[j].n1 == lastNumber) {
        /* use the new piece "straight" */
        currChain[currLength++] = j;
        pieces[j].taken = 1;
        stop = 0;
        bestLength = chain_r(currChain, currLength, bestChain, bestLength, 
                             pieces, num);
        pieces[j].taken = 0;
        currLength--;
      } else if (pieces[j].n2 == lastNumber) {
        /* use the new piece "inverted" */
        currChain[currLength++] = -j;
        pieces[j].taken = 1;
        stop = 0;
        bestLength = chain_r(currChain, currLength, bestChain, bestLength, 
                             pieces, num);
        pieces[j].taken = 0;
        currLength--;
      }
    }
  }

  if (stop && currLength>bestLength) {
    /* no possible improvement, but better solution found */
    bestLength = currLength;
    for (j=0; j<currLength; j++) {
      bestChain[j] = currChain[j];
    }
  }
  return bestLength;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
