/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 07 - exercise 12 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* structure declaration */
typedef struct {
  int v;
  int f;
} hitori_t;

/* function prototypes */
hitori_t **read_file (char *, int *);
int find_solution_r (hitori_t **, int, int, int, int);
int check_solution (hitori_t **, int);
void print_solution (char *, hitori_t **, int);

/*
 *  main program
 */
int main (int argc, char *argv[]) {
  hitori_t **hitori;
  int n, i, ok=0;

  hitori = read_file(argv[1], &n);
  fprintf(stdout, "Initial matrix:\n"); 
  print_solution("stdout", hitori, n);

  for (i=0; i<n*n && !ok; i++) {
    if (find_solution_r(hitori, n, i, 0, 0)) {
      ok = 1;
      /* output on screen */
      fprintf(stdout, "Solution (#deleted numbers = %d):\n", i); 
      print_solution("stdout", hitori, n);
      /* output on file */
      print_solution(argv[2], hitori, n);
    }
  }

  for (i=0; i<n; i++) {
    free(hitori[i]);
  }
  free(hitori);

  return EXIT_SUCCESS;
}

/* 
 *  store the input file contents
 */
hitori_t **read_file (char *name, int *nptr) {
  hitori_t **hitori;
  int n, i, j;
  FILE *fp;

  fp = fopen(name, "r");
  fscanf(fp, "%d", &n);

  hitori = (hitori_t **)malloc(n*sizeof(hitori_t *));  
  if (hitori==NULL){
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<n; i++) {
    hitori[i] = (hitori_t *)malloc(n*sizeof(hitori_t));  
    if (hitori[i]==NULL){
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    for (j=0; j<n; j++) {
      fscanf(fp, "%d", &hitori[i][j].v); 
      hitori[i][j].f = 1; 
    }
  }
  fclose(fp);

  *nptr = n;
  return hitori;
}

/*
 *  erase n numbers from the hitori matrix, recursive function
 */
int find_solution_r (hitori_t **hitori, int n, int level, int r, int c) {
  int i, j;

  if (level == 0) {
    return check_solution(hitori, n);
  }

  /* erase another number from the previous erased position on */
  for (i=r; i<n; i++) {
    for (j=(i==r)?c:0; j<n; j++) {
      if (hitori[i][j].f != -1) {
        /* erase one number */
        hitori[i][j].f = -1;
        if (find_solution_r(hitori, n, level-1, i, j)) {
          return 1;
        }
        hitori[i][j].f = 1;
      }
    }
  }

  return 0;
}

/*
 *  check whether the current mutrix is a solution of the hitori game
 */
int check_solution (hitori_t **hitori, int n) {
  int i, j, k;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      if (hitori[i][j].f != -1) {
        /* check row */
        for (k=0; k<n; k++) {
          if (k!=j && hitori[i][k].f!=-1 && hitori[i][j].v==hitori[i][k].v) {
            return 0;
          }
        }
        /* check column */
        for (k=0; k<n; k++) {
          if (k!=i && hitori[k][j].f!=-1 && hitori[i][j].v==hitori[k][j].v) {
            return 0;
          }
        }
      }
    }
  }

  return 1;
}

/*
 *  print the hitori solution on file (possibly on screen)
 */
void print_solution (char *name, hitori_t **hitori, int n) {
  FILE *fp;
  int r, c;
 
  if (strcmp(name, "stdout") == 0) {
    fp = stdout;
  } else {
    fp = fopen(name, "w");
  }

  for (r=0; r<n; r++) {
    for (c=0; c<n; c++) {
      if (hitori[r][c].f == -1) {
        fprintf(fp, "    "); 
      } else {
        fprintf(fp, "%3d ", hitori[r][c].v); 
      }
    }
    fprintf(fp, "\n"); 
  }

  if (strcmp(name, "stdout") != 0) {
    fclose(fp);
  }
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
