/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 07 - exercise 14 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define NUM   2

/* structure declaration */
typedef struct {
  char *word;
  int n;
} string;

/* macro definition */
#define strend(word)            ((word) + strlen(word) - NUM)
#define strmid(word)            ((word) + NUM)
#define chainable(word1, word2) (strncmp(strend(word1), word2, NUM)==0)

/* function prototypes */
string *read_file (int *, int);
int chain_r (string *, int, int *, int, int *, int, int);
void print_result (string *, int *, int);

/*
 *  main program
 */
int main (void) {
  int i, dim, nrip, *chainBest, *chainCurr, length;
  string *array;

  fprintf(stdout, "Input number of allowed repetitions: ");
  scanf("%d", &nrip);
  array = read_file(&dim, nrip); 

  /* allocate temporary arrays (current and chainBest solution) */
  chainCurr = (int *)malloc((dim*nrip + 1) * sizeof(int));
  chainBest = (int *)malloc((dim*nrip + 1) * sizeof(int));
  if (chainCurr==NULL || chainBest==NULL){
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<=dim*nrip; i++) {
    chainCurr[i] = chainBest[i] = -1;
  }
   
  /* search solution */
  length = chain_r(array, dim, chainCurr, NUM, chainBest, NUM, 0);

  /* output result */
  print_result(array, chainBest, length);

  /* free memory */
  for (i=0; i<dim; i++) {
    free(array[i].word);
  }
  free(array);
  free(chainBest);
  free(chainCurr);

  return EXIT_SUCCESS;
}

/* 
 *  load the input file
 */
string *read_file (int *dim_ptr, int nrip) {
  string *array;
  char word[31];
  int dim, i;
  FILE *fp;

  fprintf(stdout, "Input file name: ");
  scanf("%s", word);
  fp = fopen(word, "r");
  fscanf(fp, "%d", &dim);
  array = (string *)malloc(dim * sizeof(string));
  if (array==NULL){
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<dim; i++) {
    fscanf(fp, "%s", word);
    array[i].word = strdup(word);
    array[i].n = nrip;
  }

  fclose(fp);
  *dim_ptr = dim;
  return array;
}

/* 
 *  build the longest string that can be generated, recursive function
 */
int chain_r (
  string *array, int dim, int *chainCurr, int lengthCurr,
  int *chainBest, int lengthBest, int level
) {
  int i, j, last;

  last = (level == 0) ? -1 : chainCurr[level-1];
  for (i=0; i<dim; i++) {
    if (last==-1 || (array[i].n>0 && chainable(array[last].word, array[i].word))) {
      chainCurr[level] = i;
      array[i].n--;
      lengthCurr += strlen(strmid(array[i].word));
      if (lengthCurr > lengthBest) {
        lengthBest = lengthCurr;
        for (j=0; j<=level; j++) {
          chainBest[j] = chainCurr[j];
        }
        chainBest[level+1] = -1;
      }

      lengthBest = chain_r(array, dim, chainCurr, lengthCurr, 
                               chainBest, lengthBest, level+1);

      chainCurr[level] = -1;
      array[i].n++;
      lengthCurr -= strlen(strmid(array[i].word));
    }
  }

  return lengthBest;
}

/*
 *  display the result
 */
void print_result (string *array, int *chainBest, int lengthBest) {
  int i, j;

  fprintf(stdout, "Longest string (%d chars):\n", lengthBest);
  fprintf(stdout, "%s", array[chainBest[0]].word);
  i = 1;
  while (chainBest[i] != -1) {
    j = chainBest[i++];
    fprintf(stdout, "%s", strmid(array[j].word));
  }
  fprintf(stdout, "\n");

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
