/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 09 - exercise 03 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include "queuePrivate.h"
#include "util.h"

/*
 *  create a new empty queue
 */
queue_t *queue_init(int size) {
  queue_t *qp;

  qp = (queue_t *)util_malloc(sizeof(queue_t));
  qp->size = size;
  qp->head = qp->tail = qp->num = 0;
  qp->array = (void **)util_malloc(size*sizeof(void *));
  return qp;
}

/*
 *  return the number of elements stored in the queue
 */
int queue_count(queue_t *qp) {
  return (qp!=NULL) ? qp->num : 0;
}

/*
 *  store a new value in the queue (FIFO policy)
 */
int queue_put(queue_t *qp, void *data) {
  if (qp==NULL || qp->num>=qp->size) {
    return 0;
  }

  qp->array[qp->tail] = data;
  qp->tail = (qp->tail+1) % (qp->size);
  qp->num++;
  return 1;
}

/*
 *  extract a value from the queue (FIFO policy)
 */
int queue_get(queue_t *qp, void **data_ptr) {
  if (qp==NULL || qp->num<=0) {
    return 0;
  }

  *data_ptr = qp->array[qp->head];
  qp->head = (qp->head+1) % (qp->size);
  qp->num--;
  return 1;
}

/*
 *  print all the queue elements (FIFO policy)
 */
void queue_print(FILE *fp, queue_t *qp, void (*print)(FILE *, void *)) {
  int i, j;

  if (qp != NULL) {
    for (i=0; i<qp->num; i++) {
      j = (qp->head+i) % (qp->size);
      print(fp, qp->array[j]);
      fprintf(fp, "\n");
    }
  }
}

/*
 *  deallocate all the memory associated to the queue
 */
void queue_dispose(queue_t *qp, void (*quit)(void *)) {
  int i;

  if (qp != NULL) {
    if (quit != NULL) {
      for (i=qp->head; qp->num>0; i=(i+1)%qp->size) {
        quit(qp->array[i]);
        qp->num--;
      }
    }
    free(qp->array);
    free(qp);
  }

  return;}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
