/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 09 - exercise 04 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include "util.h"
#include "list.h"

/* structure declaration */
struct list_s {
  void *data;
  struct list_s *next;
};

/*
 *  create a new empty list
 */
list_t *list_init (void) {
  return NULL;
}

/*
 *  return the number of elements of a list
 */
int list_count (list_t *lp) {
  int size=0;

  while (lp != NULL) {
    size++;
    lp = lp->next;
  }

  return size;
}

/*
 *  search an object in a list
 */
int list_search (
 list_t *lp, void *key, float (*compare)(void *, void *), 
 void **data_ptr
) {
  for ( ; lp!=NULL; lp=lp->next) {
    if (compare(lp->data, key) == 0) {
      *data_ptr = lp->data;
      return 1;
    }
  }

  return 0;
}

/*
 *  add an object into a list (position specified!)
 */
list_t *list_add (list_t *lp, void *data, int pos) {
  list_t *node, *ptr;
  int idx = 1;

  node = (list_t *)util_malloc(sizeof(list_t));
  node->data = data;

  /* insertion ahead */
  if (lp==NULL || pos==0) {
    node->next = lp;
    return node;
  }

  /* insertion in the middle (or at the end) */
  ptr = lp;
  while (ptr->next!=NULL && idx!=pos) {
    ptr = ptr->next;
    idx++;
  }
  node->next = ptr->next;
  ptr->next = node;

  return lp;
}

/*
 *  add an object into a sorted list
 */
list_t *list_insert (list_t *lp, void *data, float (*compare)(void *, void *)) {
  list_t *ptr, *node;

  node = (list_t *)util_malloc(sizeof(list_t));
  node->data = data;

  /* insertion ahead */
  if (lp==NULL || compare(data, lp->data)<0) {
    node->next = lp;
    return node;
  }

  /* insertion in the middle (or at the end) */
  ptr = lp; 
  while (ptr->next!=NULL && compare(ptr->next->data, data)<0) {
    ptr = ptr->next;
  }
  node->next = ptr->next;
  ptr->next = node;

  return lp;
}

/*
 *  extract a specified object from a list
 */
list_t *list_extract (
  list_t *lp, void *key, float (*compare)(void *, void *), void **data_ptr
) {
  list_t *node, *ptr;

  *data_ptr = NULL;

  if (lp == NULL) {
    return NULL;
  }
  ptr = lp;

  /* extraction ahead */
  if (compare(key, lp->data) == 0) {
    *data_ptr = lp->data;
    lp = lp->next;
    free(ptr);
    return lp;
  }

  /* extraction in the middle (or at the end) */
  while (ptr->next!=NULL && compare(key, ptr->next->data)!=0) {
    ptr = ptr->next;
  }
  if (ptr->next != NULL) {
    node = ptr->next;
    *data_ptr = node->data;
    ptr->next = node->next;
    free(node);
  }

  return lp;
}

/*
 *  write a list on file
 */
void list_print (FILE *fp, list_t *lp, void (*print)(FILE *, void *)) {
  while (lp != NULL) {
    print(fp, lp->data);
    lp = lp->next;
  }
}

/*
 *  write a list on file
 */
void list_dispose (list_t *lp, void (*quit)(void *)) {
  list_t *ptr;

  while (lp != NULL) {
    ptr = lp;
    lp = lp->next;
    if (quit != NULL) {
      quit(ptr->data);
    }
    free(ptr);
  }
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
