/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 09 - exercise 06 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include "util.h"
#include "point.h"

/* static function prototypes */
static point_t *load(int *dim);
static void sort(point_t *v, int n);
static void write(point_t *v, int n);

/*
 *  main program
 */
int main (void) {
  int n, i, j, imax, jmax, imin, jmin, cnt=0;
  float d, d2, max=0, min=FLT_MAX;
  point_t *array;

  array = load(&n);
  fprintf(stdout, "Value for the distance d? ");
  scanf("%f", &d);

  for (i=0; i<n-1; i++) {
    for (j=i+1; j<n; j++) {
      d2 = distance2(array[i], array[j]);
      if (d2 > max) {
        max = d2;
        imax = i;
        jmax = j;
      }
      if (d2 < min) {
        min = d2;
        imin = i;
        jmin = j;
      }
      if (d2 < d) {
        cnt++;
      }
    }
  }

  fprintf(stdout, "Minimum distance: %.3f\n", min);
  fprintf(stdout, "Point 1: ");
  point_print(stdout, array[imin]);
  fprintf(stdout, "Point 2: ");
  point_print(stdout, array[jmin]);

  fprintf(stdout, "Maximum distance: %.3f\n", max);
  fprintf(stdout, "Point 1: ");
  point_print(stdout, array[imax]);
  fprintf(stdout, "Point 2: ");
  point_print(stdout, array[jmax]);

  fprintf(stdout, "Number of segments with length < %.2f: %d\n", d, cnt);

  sort(array, n);
  write(array, n);

  free(array);
  
  return EXIT_SUCCESS;
}

/*
 *  load the data from file
 */
static point_t *load (int *dim) {
  char name[21];
  point_t *v, t;
  FILE *fp;
  int i, n=0;

  fprintf(stdout, "Input file name? ");
  scanf("%s", name);
  fp = util_fopen(name, "r");
  while (point_read(fp, &t) != EOF) {
    n++;
  }
  fclose(fp);

  v = (point_t *)util_malloc(n * sizeof(point_t));

  fp = util_fopen(name, "r");
  for (i=0; i<n; i++) {
    point_read(fp, &v[i]);
  }
  fclose(fp);

  *dim = n;
  fprintf(stdout, "Number of loaded points: %d\n", n);
  return v;
}

/*
 *  sort the array of points according to their distance from origin
 */
static void sort (point_t *v, int n) {
  int i, j;
  point_t key;

  /* insertion sort */
  for (i=1; i<n; i++) {
    key = v[i];
    j = i;
    while (--j>=0 && distance1(key)<distance1(v[j])) {
      v[j+1] = v[j];
    }
    v[j+1] = key;
  }
}

/*
 *  write the array of points on file
 */
static void write (point_t *v, int n) {
  char name[21];
  FILE *fp;
  int i;

  fprintf(stdout, "Output file name? ");
  scanf("%s", name);
  fp = fopen(name, "w");
  for (i=0; i<n; i++) {
    point_print(fp, v[i]);
  }
  
  fclose(fp);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
