/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 * chapter 09 - exercise 08 - solution 3
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "util.h"
#include "sort.h"

#define DEBUG  1

/* structure declarations */
typedef struct {
  int id;
  char *name;
  int exams;
  float avg;
} student_t;

/* static function prototypes */
static student_t **loadStudents(int *size_ptr);
static void printStudents(student_t **array, int num);
static int compareStudents(void *ptr1, void *ptr2);
static void disposeStudent(void *ptr);

/*
 *  main program
 */
int main (void) {
  student_t **array;
  int num;

  /* load the data array */
  array = loadStudents(&num);

  /* sort the data array */
  sort_quick((void **)array, num, compareStudents);

  /* print the resulting array */
  printStudents(array, num);

  util_array_dispose((void **)array, num, disposeStudent);
  return EXIT_SUCCESS;
}

/*
 *  load the data from file
 */
static student_t **loadStudents (int *num_ptr) {
  char filename[51], name[51];
  student_t **array, *tmp;
  int id, exams, num, i;
  float avg;
  FILE *fp;

  fprintf(stdout, "Input the file name: ");
  scanf("%s", filename);

  fp = util_fopen(filename, "r");
  fscanf(fp, "%d", &num);
  array = (student_t **)util_malloc(num*sizeof(student_t *));
  for (i=0; i<num; i++) {
    fscanf(fp, "%d %s %d %f", &id, name, &exams, &avg);
    tmp = (student_t *)util_malloc(sizeof(student_t));
    tmp->id = id;
    tmp->name = util_strdup(name);
    tmp->exams = exams;
    tmp->avg = avg;
    array[i] = tmp;
  }
  fclose(fp);

#if DEBUG
  /* debug print out */
  fprintf(stdout, "Number of data loaded: %d\n", num);
  printStudents(array, num);
  fprintf(stdout, "\n");
#endif

  *num_ptr = num;
  return array;
}

/* 
 *  print-out the sorted data array
 */
static void printStudents (student_t **array, int num) {
  int i;

  for (i=0; i<num; i++) {
    fprintf(stdout, "%7d %-20s ", array[i]->id, array[i]->name);
    fprintf(stdout, "%2d %.2f\n", array[i]->exams, array[i]->avg);
  }
}

/*
 *  compare two students (by id)
 */
static int compareStudents (void *ptr1, void *ptr2) {
  student_t *s1 = (student_t *)ptr1;
  student_t *s2 = (student_t *)ptr2;
  return s1->id - s2->id;
}

/*
 *  free a student structure
 */
static void disposeStudent (void *ptr) {
  student_t *s = (student_t *)ptr;
  free(s->name);
  free(s);
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part II: Algorithms and Data Structures
 *
 */
